/*
 * file PropertyNameList.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM
 *
 * (c) Copyright IBM Corporation 2004, 2008.  All Rights Reserved. 
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or  
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp. 
 */
package javax.wvcm;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
 * A list of resource property names.
 * 
 * @since 1.0
 */
public final class PropertyNameList implements PropertyRequestItem {

    /**
     * The name of a property of a persistent resource. The PropertyName
     * type parameter specifies the type of the value of the PropertyName
     */
    public static class PropertyName<T> implements PropertyRequestItem {

        private final String _namespace;
        private final String _name;

        /**
         * Create a PropertyName with a null namespace and the given name.
         * 
         * @param name the name for this property.  May not be null.
         */
        PropertyName(String name) {
            this(null, name);
        }

        /**
         * Create a PropertyName with the given namespace and name.
         * 
         * @param namespace the namespace for this PropertyName.
         * May be null to indicate a system property, or non-null
         * to indicate a property in the specified name space.
         * @param name the name for this property.  May not be null.
         */
        public PropertyName(String namespace, String name) {
            _namespace = namespace;
            _name = name;
        }

        /**
         * Get the namespace of the property name.
         * 
         * @return the namespace.  May be null to indicate a system property,
         * or non-null to indicate a property in the specified name space.
         */
        public String getNamespace() {
            return _namespace;
        }

        /**
         * Get the property name.
         * 
         * @return the name part of the property name */
        public String getName() {
            return _name;
        }

        /**
         * Calculate a hash code value for the object.
         * 
         * @return the hash code for the object.
         */
        @Override
        public int hashCode() {
            int result = 17;
            result = 37*result + ((_namespace == null) ? 0 : _namespace.hashCode());
            result = 37*result + _name.hashCode();
            return result;
        }

        /**
         * Indicates whether some other object is "equal to" this one.
         * 
         * @param o the object to compare with.
         * @return true if and only if the specified object is a PropertyName whose
         * type, namespace value, and name value equals those of this object.
         */
        @Override
        public boolean equals(Object o) {
            if (o == null || this.getClass() != o.getClass()) {
                return false;
            }
            PropertyName<?> pn = (PropertyName<?>) o;
            if (_namespace == null) {
                return pn._namespace == null && _name.equals(pn._name);
            }
            return _namespace.equals(pn._namespace) && _name.equals(pn._name);
        }

        /**
         * Returns a string representation of the PropertyName for diagnostic 
         * purposes.  
         * @see java.lang.Object#toString
         * 
         * @return The string representation of this PropertyName in the format
         * [namespace:]name.
         */
        @Override
        public String toString() {
            if (_namespace != null)
                return _namespace + ":" + _name; //$NON-NLS-1$
            else
                return _name;
        }
        
        /**
         * Constructs a NestedPropertyName whose root property is this
         * PropertyName with the property request items supplied as arguments as
         * its nested property request. This method allows a more succinct
         * syntax for the declaration of PropertyRequest objects at compile
         * time. For example,
         * 
         * <pre>
         * final static PropertyRequest WORKSPACE_PROPERTIES = 
         *     new PropertyRequest(
         *         Workspace.DISPLAY_NAME,
         *         Workspace.COMMENT,
         *         Workspace.ACTIVITY_LIST.nest( 
         *             Activity.DISPLAY_NAME,
         *             Activity.ACTIVITY_VERSION_LIST.nest( 
         *                 Version.DISPLAY_NAME, 
         *                 Version.COMMENT, 
         *                 Version.CREATION_DATE)));
         * </pre>
         * 
         * @param pnl An array of PropertyRequestItem objects that specify the
         *            properties to be requested from the value of the property
         *            identified by this PropertyName.
         * @return A NestedPropertyName whose root is this PropertyName and
         *         whose nested properties are the property request items given
         *         as arguments.
         */
        public NestedPropertyName<T> nest(PropertyRequestItem... pnl)
        {
            return new NestedPropertyName<T>(this, pnl);
        }

        /**
         * A convenience method for constructing nested property names from a
         * NestedPropertyName[] without requiring an explicit cast to
         * PropertyRequestItem[]
         * 
         * @param pnl A NestedPropertyName[] that specifies the nested
         *            properties of the NestedPropertyName to be constructed
         * @return A NestedPropertyName whose root is this PropertyName and
         *         whose nested properties are the elements of the
         *         NestedPropertyNames[] argument.
         */
        public NestedPropertyName<T> nest(NestedPropertyName<?>[] pnl)
        {
            return new NestedPropertyName<T>(this, (PropertyRequestItem[])pnl);
        }

        /**
         * A convenience method for constructing nested property names from a
         * PropertyName[] without requiring an explicit cast to
         * PropertyRequestItem[]
         * 
         * @param pnl A PropertyName[] that specifies the nested properties of
         *            the NestedPropertyName to be constructed
         * @return A NestedPropertyName whose root is this PropertyName and
         *         whose nested properties are the elements of the
         *         PropertyNames[] argument.
         */
        public NestedPropertyName<T> nest(PropertyName<?>[] pnl)
        {
            return new NestedPropertyName<T>(this, (PropertyRequestItem[])pnl);
        }
    }

    // An empty immutable PropertyName[]
    private static final PropertyName<?>[] EMPTY_PNA = new PropertyName[0];

    // The immutable array of property names for this PropertyNameList.
    private final PropertyName<?>[] _propertyNames;

    // The lazy initialized Set for this immutable object.
    private volatile Set<PropertyName<?>> _asSet = null;

    /**
     * Construct a PropertyNameList from an array of property names.
     * 
     * @param propertyNames an array of property names, or null.
     * Passing null is equivalent to passing an empty array.
     */
    public PropertyNameList(PropertyName<?>... propertyNames) {
        _propertyNames = (propertyNames == null || propertyNames.length == 0)
        ? EMPTY_PNA : propertyNames;
    }

    /**
     * Get an array of property names from a property name list.
     * 
     * @return the array of property names maintained by this 
     * PropertyNameList. Will never be <code>null</code>.
     */
    public PropertyName<?>[] getPropertyNames() {
        return _propertyNames;
    }

    /**
     * Indicates whether some other object is "equal to" this one.
     * 
     * @param o the object to compare with.
     * @return true if and only if the specified object is a PropertyNameList
     *         whose property names array is equal to the property names array
     *         of this ignoring order.
     */
    @Override
    public boolean equals(Object o) {
        if (o == null || this.getClass() != o.getClass()) {
            return false;
        }
        PropertyNameList pnl = (PropertyNameList) o;
        return getPropertyNamesAsSet().equals(pnl.getPropertyNamesAsSet());
    }


    /**
     * Calculate a hash code for a PropertyNameList.
     * 
     * @return a hash code for the list.
     */
    @Override
    public int hashCode() {
        return getPropertyNamesAsSet().hashCode();
    }

    /**
     * Returns a string representation of this PropertyNameList suitable for
     * diagnostics.
     * 
     * @return The string representation of this PropertyNameList formatted as
     * "[&LT;first property name&GT;, &LT;second property name&GT;]"
     */
    @Override
    public String toString() {

        StringBuffer builder = new StringBuffer();
        builder.append('[');

        for(PropertyName<?> name: getPropertyNames()) {
            if (builder.length() > 1)
                builder.append(", "); //$NON-NLS-1$
            
            builder.append(name.toString());
        }
            
        builder.append(']');

        return builder.toString();
    }

    // the property names maintained by this PropertyNameList as a Set.
    private Set<PropertyName<?>> getPropertyNamesAsSet() {
        if (_asSet == null) {
            if (_propertyNames == EMPTY_PNA)
                _asSet = Collections.emptySet();
            else
                _asSet = new HashSet<PropertyName<?>>(Arrays.asList(_propertyNames));
        }
        return _asSet;
    }
}
